from typing import List, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class UnpivotStep(BaseStep):
    name = Field('unpivot', const=True)
    keep: List[ColumnName]
    unpivot: List[ColumnName]
    unpivot_column_name: ColumnName
    value_column_name: ColumnName
    dropna: bool


class UnpivotStepWithVariable(UnpivotStep, StepWithVariablesMixin):
    keep: Union[TemplatedVariable, List[TemplatedVariable]]
    unpivot: Union[TemplatedVariable, List[TemplatedVariable]]



from distutils import log
from typing import List

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import SortStep
from weaverbird.pipeline.steps.sort import ColumnSort


def sort_columns_to_sql(columns: List[ColumnSort]) -> str:
    """
    Concatenate fields and order's fields

    params:
    columns: List[ColumnSort]
    """
    compiled_query: str = ''
    for index, c in enumerate(columns):
        if index > 0:
            compiled_query += ', '
        compiled_query += f'{c.column} {c.order}'

    return compiled_query


def translate_sort(
    step: SortStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'SORT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step.columns: {step.columns}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str()

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields} FROM {query.query_name}"""
        f""" ORDER BY {sort_columns_to_sql(step.columns)})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Tuple

from weaverbird.backends.sql_translator.types import SQLQuery, SQLStepTranslationReport
from weaverbird.pipeline import Pipeline, PipelineStep

from .steps import sql_steps_translators
from .types import (
    SQLPipelineTranslationReport,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)


def translate_pipeline(
    pipeline_to_translate: Pipeline,
    sql_query_retriever: SQLQueryRetriever,  # TODO : rename it to sql_table_retriever when ready
    sql_query_describer: SQLQueryDescriber,
    sql_query_executor: SQLQueryExecutor,
    subcall_from_other_pipeline_count: int = None,
) -> Tuple[SQLQuery, SQLPipelineTranslationReport]:
    """
    The main function of the module. Translates a pipeline and returns the result as a transformed query.
    For example, a Select step chained with a filter step will output a query string like this:

    with select_step_1 as (select * from "DB"."SCHEMA"."TABLE" limit 100),
    filter_step_2 as (select * from select_step_1 where ID > 2)
    select * from filter_step_2

    """
    query = SQLQuery()
    translate_report = []
    for index, step in enumerate(pipeline_to_translate.steps):
        try:
            query = sql_steps_translators[step.name](
                step,
                query,
                index,
                sql_query_retriever=sql_query_retriever,
                sql_query_describer=sql_query_describer,
                sql_query_executor=sql_query_executor,
                sql_translate_pipeline=translate_pipeline,
                subcall_from_other_pipeline_count=subcall_from_other_pipeline_count,
            )
            translate_report.append(SQLStepTranslationReport(step_index=index))

        except Exception as e:
            raise SQLPipelineTranslationFailure(step, index, e) from e

    query_string = f'{query.transformed_query} {query.selection_query}'
    return query_string, SQLPipelineTranslationReport(
        sql_steps_translation_reports=translate_report
    )


class SQLPipelineTranslationFailure(Exception):
    """Raised when a error happens during the translation of the pipeline"""

    def __init__(self, step: PipelineStep, index: int, original_exception: Exception):
        self.step = step
        self.index = index
        self.original_exception = original_exception
        self.message = f'Step #{index + 1} ({step.name}) failed: {original_exception}'
        self.details = {'index': index, 'message': self.message}



from typing import List, Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ConvertStep(BaseStep):
    name = Field('convert', const=True)
    columns: List[ColumnName]
    data_type: Literal['integer', 'float', 'text', 'date', 'boolean']



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import CumSumStep


def translate_cumsum(
    step: CumSumStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'CUMSUM_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    # if any new column name had been provided
    if step.new_column is None:
        step.new_column = f"{step.value_column}_CUMSUM"

    # we complete fields without the new column name
    # if there is a same naming
    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=[step.new_column]
    )

    # the partition by sub query
    partition_by_sub_query = 'NULL'
    # if there is a group by query
    if step.groupby is not None and len(step.groupby):
        partition_by_sub_query = ', '.join(step.groupby)

    # The final query
    final_query = (
        f"SELECT {completed_fields}, SUM({step.value_column}) OVER (PARTITION BY {partition_by_sub_query}"
        f" ORDER BY {step.reference_column} ASC rows UNBOUNDED PRECEDING) {step.new_column}"
        f" FROM {query.query_name} ORDER BY {step.reference_column} ASC"
    )

    # we make sure to add the new column in the metadata list
    query.metadata_manager.add_query_metadata_column(step.new_column, "FLOAT")

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS ({final_query})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import EvolutionStep


def translate_evolution(
    step: EvolutionStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    """
    Example of generated query (evolution on date column + 1 year, groupby orderstatus):
    SELECT A.O_TOTALPRICE AS P1, B.O_TOTALPRICE AS P2, A.O_ORDERDATE AS DATE, A.O_ORDERSTATUS as STAT,
    B.O_ORDERSTATUS AS STATB
    FROM ORDERS_TRUNC A
    LEFT JOIN ORDERS_TRUNC B
    ON A.O_ORDERDATE = DATEADD('year', 1, B.O_ORDERDATE) AND A.O_ORDERSTATUS = B.O_ORDERSTATUS
    WHERE DATEADD('year', 1, B.O_ORDERDATE) IS NULL OR B.O_ORDERSTATUS IS NULL
    """

    DATE_UNIT = {
        'vsLastYear': 'year',
        'vsLastMonth': 'month',
        'vsLastWeek': 'week',
        'vsLastDay': 'day',
    }

    query_name = f'EVOLUTION_STEP_{index}'
    new_column_name = (
        step.new_column.upper()
        if step.new_column
        else f'{step.value_col}_EVOL_{step.evolution_format.upper()}'
    )

    if step.evolution_format == 'abs':
        new_column = f'''(A.{step.value_col} - B.{step.value_col}) AS {new_column_name}'''
    else:
        new_column = f'''((A.{step.value_col} / B.{step.value_col}) - 1) AS {new_column_name}'''

    selected_columns = [
        f'A.{c} AS {c}' for c in query.metadata_manager.retrieve_query_metadata_columns()
    ] + [new_column]

    date_join = (
        f"A.{step.date_col} = DATEADD('{DATE_UNIT[step.evolution_type]}', 1, B.{step.date_col})"
    )

    on_query = f"ON {' AND '.join([date_join]+[f'A.{c} = B.{c}' for c in step.index_columns])}"

    final_query = (
        f"SELECT {', '.join(selected_columns)}"
        f" FROM {query.query_name} A LEFT JOIN {query.query_name} B"
        f" {on_query} ORDER BY A.{step.date_col}"
    )

    transformed_query = f"{query.transformed_query}, {query_name} AS ({final_query})"

    query.metadata_manager.add_query_metadata_column(new_column_name, 'float')

    return SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )



from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin


class CustomSqlStep(BaseStep):
    name = Field('customsql', const=True)
    query: str


class CustomSqlStepWithVariables(CustomSqlStep, StepWithVariablesMixin):
    name = Field('customsql', const=True)
    query: str



from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_hierarchical_columns_list,
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import RollupStep


def translate_rollup(
    step: RollupStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:

    query_name = f'ROLLUP_STEP_{index}'
    all_columns = build_hierarchical_columns_list(step)
    group_by_part = f"GROUP BY {', '.join(step.groupby)}, " if step.groupby else 'GROUP BY '
    rollup_part = f"ROLLUP({', '.join(step.hierarchy)}) HAVING {step.hierarchy[0]} IS NOT NULL"
    transformed_query = f"""{query.transformed_query}, {query_name} AS (SELECT {all_columns} FROM {query.query_name} {group_by_part}{rollup_part})"""
    query.metadata_manager.remove_query_metadata_all_columns()
    query.metadata_manager.add_query_metadata_columns(
        sql_query_describer(
            domain=None, query_string=f'{transformed_query} SELECT * FROM {query_name}'
        )
    )
    return SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )



from distutils import log
from typing import List, Sequence

from weaverbird.backends.sql_translator.metadata import ColumnMetadata, SqlQueryMetadataManager
from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import TotalsStep
from weaverbird.pipeline.steps.aggregate import Aggregation
from weaverbird.pipeline.steps.totals import TotalDimension
from weaverbird.pipeline.types import ColumnName
from weaverbird.utils.iter import combinations


def make_totals_query(step: TotalsStep, parent_query: SQLQuery) -> (str, List[ColumnMetadata]):
    def select_total_dimensions(total_dimensions: List[TotalDimension]) -> List[ColumnMetadata]:
        selects = []
        for dimension in total_dimensions:
            selects.append(
                # in practice, it should be something called a `SelectField` if our queries were typed objects.
                # however this `ColumnMetadata` type has all I need.
                ColumnMetadata(
                    name=f"CASE WHEN GROUPING({dimension.total_column}) = 0 THEN {dimension.total_column} ELSE '{dimension.total_rows_label}' END",
                    alias=dimension.total_column,
                    type=parent_query.metadata_manager.retrieve_query_metadata_column_type_by_name(
                        dimension.total_column
                    ),
                )
            )
        return selects

    def select_aggregate(aggregations: Sequence[Aggregation]) -> List[ColumnMetadata]:
        aggregated_cols = []
        for aggregation in [
            aggregation
            for aggregation in aggregations
            if aggregation.agg_function not in ['first', 'last']
        ]:
            for col, new_col in zip(aggregation.columns, aggregation.new_columns):
                if aggregation.agg_function == 'count distinct':
                    aggregated_cols.append(
                        ColumnMetadata(name=f'COUNT(DISTINCT {col})', alias=new_col, type='FLOAT')
                    )
                else:
                    aggregated_cols.append(
                        ColumnMetadata(
                            name=f'{aggregation.agg_function.upper()}({col})',
                            alias=new_col,
                            type='FLOAT',
                        )
                    )
        return aggregated_cols

    def with_alias(selects: List[ColumnMetadata]) -> List[str]:
        selects_with_alias = []
        for select in selects:
            select_str = select.original_name
            if select.alias is not None:
                select_str = f'{select_str} AS "{select.alias}"'
            selects_with_alias.append(select_str)
        return selects_with_alias

    total_dimensions: List[ColumnName] = list(map(lambda x: x.total_column, step.total_dimensions))

    selects: List[ColumnMetadata] = (
        select_total_dimensions(step.total_dimensions)
        + select_aggregate(step.aggregations)
        + [
            ColumnMetadata(
                name=column,
                type=parent_query.metadata_manager.retrieve_query_metadata_column_type_by_name(
                    column
                ),
            )
            for column in step.groups
        ]
    )

    group_sets = []
    for combination in combinations(total_dimensions):
        group_sets.append('(' + ', '.join(combination) + ')')
    group_sets.append('()')
    group_by = step.groups + ['GROUPING SETS(' + (', '.join(group_sets)) + ')']

    query = f"""SELECT {', '.join(with_alias(selects))} FROM {parent_query.query_name} GROUP BY {', '.join(group_by)}"""

    return query, selects


def translate_totals(
    step: TotalsStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'TOTALS_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    sql_query, selects = make_totals_query(step, query)
    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"{query.transformed_query}, {query_name} AS ({sql_query})",
    )

    new_query.metadata_manager = SqlQueryMetadataManager()
    [
        new_query.metadata_manager.add_query_metadata_column(
            select.alias if select.alias is not None else select.name, select.type
        )
        for select in selects
    ]

    new_query.selection_query = build_selection_query(
        new_query.metadata_manager.retrieve_query_metadata_columns(), query_name
    )
    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SqlQueryMetadataManager,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import TableStep


def translate_table(
    step: TableStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    """As it is always the first step add the with keyword"""
    select_from_table = sql_query_retriever(
        step.domain
    )  # TODO in laputa, implement the table retrieval instead of query
    query_name = f'SELECT_STEP_{index}'
    query_name += (
        f'_{subcall_from_other_pipeline_count}'
        if isinstance(subcall_from_other_pipeline_count, int)
        and subcall_from_other_pipeline_count >= 0
        else ''
    )
    query_description = sql_query_describer(step.domain)
    query_metadata_manager = SqlQueryMetadataManager(
        tables_metadata={step.domain: query_description}
    )

    sql_query = SQLQuery(
        query_name=query_name,
        transformed_query=f'WITH {query_name} AS ({select_from_table})',
        selection_query=build_selection_query(
            query_metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query_metadata_manager,
    )
    # TODO adapt Laputa to directly return the SQLQuery
    return sql_query



from weaverbird.backends.sql_translator.steps.utils.aggregation import (
    build_first_or_last_aggregation,
    prepare_aggregation_query,
)
from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import AggregateStep


def translate_aggregate(
    step: AggregateStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'AGGREGATE_STEP_{index}'

    aggregated_cols = []
    aggregated_string = ''
    first_last_string = ''

    query, aggregated_string, new_as_columns = prepare_aggregation_query(
        query_name, aggregated_cols, aggregated_string, query, step
    )

    query, query_string = build_first_or_last_aggregation(
        aggregated_string, first_last_string, query, step, query_name, new_as_columns
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f'{query.transformed_query}, {query_name} AS ({query_string})',
    )

    new_query.metadata_manager = query.metadata_manager
    new_query.selection_query = build_selection_query(
        query.metadata_manager.retrieve_query_metadata_columns(), query_name
    )

    return new_query



from typing import Optional

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ToDateStep(BaseStep):
    name = Field('todate', const=True)
    column: ColumnName
    format: Optional[str]



from __future__ import annotations

import copy
from typing import Dict, List, Optional

from pydantic import BaseModel


class MetadataError(Exception):
    ...


class ColumnMetadata(BaseModel):
    name: str
    original_name: Optional[str]
    type: str
    original_type: Optional[str]
    alias: Optional[str]

    delete: Optional[bool] = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'name' not in kwargs or 'type' not in kwargs:
            raise MetadataError('prout2')
        self.name: str = kwargs['name'].upper()
        self.original_name: str = kwargs['name']
        self.type: str = kwargs['type'].upper()
        self.original_type: str = kwargs['type']

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def update_name(self, dest_name: str) -> ColumnMetadata:
        self.name = dest_name.upper()
        return self

    def update_type(self, dest_type: str) -> ColumnMetadata:
        self.type = dest_type.upper()
        return self

    def update(self, column: ColumnMetadata) -> ColumnMetadata:
        self.name = column.name.upper()
        self.type = column.type.upper()
        return self

    def remove(self) -> ColumnMetadata:
        self.delete = True
        return self

    def is_deleted(self):
        return self.delete

    def update_alias(self, alias) -> ColumnMetadata:
        self.alias = alias.upper()
        return self

    def remove_alias(self) -> ColumnMetadata:
        self.alias = None
        return self


class TableMetadata(BaseModel):
    name: str
    original_name: Optional[str]

    delete: Optional[bool] = False

    columns: Optional[Dict[str, ColumnMetadata]] = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'name' not in kwargs:
            raise MetadataError('prout')
        self.name: str = kwargs['name'].upper()
        self.original_name: str = kwargs['name']

    def __eq__(self, other):
        return self.name == other.name

    def update_name(self, table_name: str) -> TableMetadata:
        self.name = table_name.upper()
        return self

    def retrieve_column_by_column_name(
        self, column_name: str, deleted: bool = False
    ) -> Optional[ColumnMetadata]:
        c_name = column_name.upper()
        if c_name in self.columns and (
            (deleted and self.columns[c_name].is_deleted()) or not self.columns[c_name].is_deleted()
        ):
            return self.columns[c_name]
        return None

    def retrieve_columns(self, deleted: bool = False) -> Dict[str, ColumnMetadata]:
        l: Dict[str, ColumnMetadata] = {}
        for c_name in self.columns:
            if (deleted and self.columns[c_name].is_deleted()) or not self.columns[
                c_name
            ].is_deleted():
                l[c_name] = self.columns[c_name]
        return l

    def retrieve_column_by_type(
        self, column_type: str, deleted: bool = False
    ) -> Dict[str, ColumnMetadata]:
        c_type = column_type.upper()
        result: Dict[str, ColumnMetadata] = {}
        for c_name, column in self.columns:
            if (
                column.type == c_type
                and (deleted and self.columns[c_name].is_deleted())
                or not self.columns[c_name].is_deleted()
            ):
                result[c_name] = column
        return result

    def add_column_o(self, column: ColumnMetadata) -> TableMetadata:
        c_name = column.name.upper()
        self.columns[c_name] = column
        return self

    def add_columns_o(self, columns: List[ColumnMetadata]) -> TableMetadata:
        for c in columns:
            self.add_column_o(column=c)
        return self

    def add_column(
        self, column_name: str, column_type: str, alias: Optional[str] = None
    ) -> TableMetadata:
        c = ColumnMetadata(name=column_name, type=column_type, alias=alias)
        self.add_column_o(c)
        return self

    def add_columns(self, columns: Dict[str, str]) -> TableMetadata:
        for k, v in columns.items():
            self.add_column(column_name=k, column_type=v)
        return self

    def update_column_name(self, column_name: str, dest_column_name: str) -> TableMetadata:
        c_name = column_name.upper()
        if c_name not in self.columns:
            raise MetadataError(f'Error to update column {c_name}({column_name}), column not exist')
        result = self.columns[c_name].update_name(dest_column_name)
        del self.columns[c_name]
        self.columns[result.name] = result
        return self

    def update_column_type(self, column_name: str, dest_column_type: str) -> ColumnMetadata:
        c_name = column_name.upper()
        if c_name not in self.columns:
            raise MetadataError(f'Error to update column {c_name}({column_name}), column not exist')
        self.columns[c_name] = self.columns[c_name].update_type(dest_column_type)
        return self.columns[c_name]

    def update_column_alias(self, column_name, alias) -> ColumnMetadata:
        c_name = column_name.upper()
        if c_name not in self.columns:
            raise MetadataError(
                f'Error updating column {c_name}({column_name}), column does not exist'
            )
        self.columns[c_name] = self.columns[c_name].update_alias(alias)
        return self

    def remove_column(self, column_name: str) -> None:
        c_name = column_name.upper()
        if c_name not in self.columns:
            raise MetadataError(f'Error to delete column {c_name}({column_name}), column not exist')
        self.columns[c_name].remove()

    def remove_columns(self, columns_name: List[str]) -> int:
        i = 0
        for name in columns_name:
            c_name = name.upper()
            if c_name not in self.columns:
                raise MetadataError(f'Error to delete column {c_name}({name}), column not exist')
            self.columns[c_name].remove()
            i += 1
        return i

    def remove_all_columns(self) -> int:
        i = len(self.columns)
        self.columns = {}
        return i

    def remove(self) -> TableMetadata:
        self.delete = True
        return self

    def remove_column_alias(self, column_name: str) -> TableMetadata:
        if column_name not in self.columns:
            raise MetadataError(f'Error updating column {column_name}), column does not exist')
        self.columns[column_name].remove_alias()
        return self

    def is_deleted(self) -> bool:
        return self.delete


class SqlQueryMetadataManager(BaseModel):
    tables: Optional[Dict[str, TableMetadata]]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.tables = {'__INTERNAL__': TableMetadata(name='__INTERNAL__')}
        if 'tables_metadata' in kwargs:
            first = True
            for table_name in kwargs['tables_metadata']:
                table = self.create_table(table_name)
                for column in kwargs['tables_metadata'][table_name]:
                    table.add_column(column, kwargs['tables_metadata'][table_name][column])
                if first:
                    self.define_as_metadata(table_name)
                    first = False

    def _check_name(self, name) -> bool:
        if name == '__INTERNAL__':
            raise MetadataError('Table __INTERNAL__ is reserved')
        return True

    def define_as_metadata(self, table_name: str) -> SqlQueryMetadataManager:
        t_name = table_name.upper()
        if t_name not in self.tables and self._check_name(t_name):
            raise MetadataError(
                f'Impossible to define table {t_name}({table_name}) - Table does not exist'
            )
        self.tables['__INTERNAL__'] = copy.deepcopy(self.tables[t_name])
        self.tables['__INTERNAL__'].original_name = table_name
        return self

    def create_table(self, table_name: str) -> TableMetadata:
        t_name = table_name.upper()

        # just a quick fix for the join, not good for now
        # TODO : Find a better way to manage the behaviour of join two tables with the same name
        # if t_name in self.tables and self._check_name(t_name):
        #     raise MetadataError(f'Table {t_name}({table_name}) already exist')

        self.tables[t_name] = TableMetadata(name=table_name)
        return self.tables[t_name]

    def remove_table(self, table_name: str) -> SqlQueryMetadataManager:
        t_name = table_name.upper()
        if t_name not in self.tables and self._check_name(t_name):
            raise MetadataError(
                f'Impossible to remove table {t_name}({table_name}) - Table does not exist'
            )
        self.tables[t_name] = self.tables[t_name].remove()
        return self

    def remove_tables(self) -> SqlQueryMetadataManager:
        self.tables = {}
        return self

    def update_table(self, table_name: str, dest_table_name: str) -> SqlQueryMetadataManager:
        t_name = table_name.upper()
        if t_name not in self.tables and self._check_name(t_name):
            raise MetadataError(f'Table {t_name}({table_name}) not exist')
        result = self.tables[t_name].update_name(dest_table_name)
        del self.tables[t_name]
        self.tables[result.name] = result
        return self

    def duplicate_table(
        self, src_table: TableMetadata, dst_table_name: str
    ) -> SqlQueryMetadataManager:
        t_name = dst_table_name.upper()

        # just a quick fix for the join, not good for now
        # TODO : Find a better way to manage the behaviour of join two tables with the same name
        # if dst_table_name in self.tables:
        #     raise MetadataError(f'Table {t_name}({dst_table_name}) already exist')

        self.create_table(t_name)
        self.tables[t_name] = copy.deepcopy(src_table)
        return self

    def join_query_metadata(self, join_table: str, left_query_name: str) -> SqlQueryMetadataManager:
        jt_name: str = join_table.upper()
        if jt_name not in self.tables:
            raise MetadataError(
                f'Impossible to merge table {jt_name}({join_table}) - Table not exist'
            )
        self.create_table(table_name='__TEMP_JOIN__')

        for cname in self.retrieve_query_metadata_columns():
            self.add_table_column(
                table_name='__TEMP_JOIN__',
                column_name=f'{left_query_name}.{cname}',
                column_type=self.retrieve_query_metadata_column_by_name(cname).type,
                alias=f'{cname}_LEFT',
            )
        self.define_as_metadata(table_name='__TEMP_JOIN__')

        for cname in self.tables[jt_name].columns:
            self.add_query_metadata_column(
                column_name=f'{jt_name}.{cname}',
                column_type=self.retrieve_column_by_name(
                    table_name=jt_name, column_name=cname
                ).type,
                alias=f'{cname}_RIGHT',
            )

        return self

    def append_queries_metadata(self, unioned_tables: List[str]) -> SqlQueryMetadataManager:
        try:
            all_columns = [self.retrieve_columns_as_list(t) for t in unioned_tables]
            all_columns.sort(key=len)
            max_column_number = max([len(cl) for cl in all_columns])
        except ValueError:
            raise MetadataError('Impossible to append tables - Table does not exist')

        new_table = TableMetadata(name='__INTERNAL__')
        new_table.original_name = self.tables['__INTERNAL__'].original_name
        new_table.columns = self.retrieve_query_metadata_columns()
        if len(new_table.columns) < max_column_number:
            index_cols_to_add = range(len(new_table.columns), max_column_number)
            for cols in all_columns:
                for index in index_cols_to_add:
                    if index < len(cols):
                        new_table.add_column(
                            column_name=f'NULL AS {cols[index]}', column_type='UNDEFINED'
                        )
        self.tables['__INTERNAL__'] = new_table
        return self

    def add_table_column(
        self, table_name: str, column_name: str, column_type: str, alias: Optional[str] = None
    ) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to update column name, table {t_name}({table_name}) not exist'
            )
        self.tables[t_name] = self.tables[t_name].add_column(column_name, column_type, alias)
        return self.tables[t_name]

    def add_table_columns_from_dict(self, table_name: str, columns_dict: Dict[str, str]):
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to update column name, table {t_name}({table_name}) does not exist'
            )
        for name, type in columns_dict.items():
            self.tables[t_name] = self.tables[t_name].add_column(name, type)

    def add_query_metadata_column(
        self, column_name: str, column_type: str, alias: Optional[str] = None
    ) -> TableMetadata:
        return self.add_table_column('__INTERNAL__', column_name, column_type, alias)

    def add_query_metadata_columns(self, columns: Dict[str, str]) -> TableMetadata:
        return self.add_table_columns_from_dict('__INTERNAL__', columns)

    def update_column_name(
        self, table_name: str, column_name: str, dest_column_name: str
    ) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to update column name, table {t_name}({table_name}) not exist'
            )
        self.tables[t_name] = self.tables[t_name].update_column_name(column_name, dest_column_name)
        return self.tables[t_name]

    def update_column_alias(self, table_name: str, column_name: str, alias: str) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to update column\'s alias, table {t_name}({table_name}) does not exist'
            )
        self.tables[t_name] = self.tables[t_name].update_column_alias(column_name, alias)
        return self.tables[t_name]

    def remove_column_alias(self, table_name: str, column_name: str) -> TableMetadata:
        if table_name not in self.tables:
            raise MetadataError(
                f'Impossible to remove column\'s alias, table {table_name} does not exist'
            )
        self.tables[table_name] = self.tables[table_name].remove_column_alias(column_name)
        return self.tables[table_name]

    def update_query_metadata_column_name(
        self, column_name: str, dest_column_name: str
    ) -> TableMetadata:
        self.update_column_name('__INTERNAL__', column_name, dest_column_name)
        return self.retrieve_table('__INTERNAL__')

    def update_query_metadata_column_alias(self, column_name: str, alias: str):
        self.update_column_alias('__INTERNAL__', column_name, alias)
        return self.retrieve_table('__INTERNAL__')

    def update_column_type(
        self, table_name: str, column_name: str, dest_column_type: str
    ) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to update column type, table {t_name}({table_name}) not exist'
            )
        self.tables[t_name].update_column_type(column_name, dest_column_type)
        return self.retrieve_table('__INTERNAL__')

    def update_query_metadata_column_type(
        self, column_name: str, dest_column_type: str
    ) -> TableMetadata:
        self.update_column_type('__INTERNAL__', column_name, dest_column_type)
        return self.retrieve_table('__INTERNAL__')

    def remove_table_column(self, table_name: str, column_name: str) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to delete column, table {t_name}({table_name}) does not exist'
            )
        self.tables[t_name].remove_column(column_name)
        return self.tables[t_name]

    def remove_query_metadata_column(self, column_name: str) -> TableMetadata:
        return self.remove_table_column('__INTERNAL__', column_name)

    def remove_query_metadata_column_alias(self, column_name: str) -> TableMetadata:
        return self.remove_column_alias('__INTERNAL__', column_name)

    def remove_table_columns(self, table_name: str, columns_name: List[str]) -> TableMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to delete column, table {t_name}({table_name}) does not exist'
            )
        self.tables[t_name].remove_columns(columns_name)
        return self.tables[t_name]

    def remove_query_metadata_columns(self, columns_name: List[str]) -> TableMetadata:
        return self.remove_table_columns('__INTERNAL__', columns_name)

    def remove_table_all_columns(self, table_name: str) -> int:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to delete column, table {t_name}({table_name}) does not exist'
            )
        return self.tables[t_name].remove_all_columns()

    def remove_query_metadata_all_columns(self) -> int:
        return self.remove_table_all_columns('__INTERNAL__')

    def retrieve_table(self, table_name: str, deleted: bool = False) -> Optional[TableMetadata]:
        t_name = table_name.upper()
        if t_name in self.tables and (
            (deleted and self.tables[t_name].is_deleted()) or not self.tables[t_name].is_deleted()
        ):
            return self.tables[t_name]
        return None

    def retrieve_query_metadata(self) -> TableMetadata:
        return self.retrieve_table('__INTERNAL__')

    def retrieve_table_columns(
        self, table_name: str, deleted: bool = False
    ) -> Dict[str, ColumnMetadata]:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to retrieve column for table {t_name}({table_name}) - Table does not exist'
            )
        return self.tables[t_name].retrieve_columns(deleted)

    def retrieve_query_metadata_columns(self, deleted: bool = False) -> Dict[str, ColumnMetadata]:
        return self.retrieve_table_columns('__INTERNAL__', deleted)

    def retrieve_columns_as_list(
        self, table_name: str, columns_filter: List[str] = None, deleted: bool = False
    ) -> List[str]:
        columns = self.retrieve_table_columns(table_name, deleted)
        filter_format = []
        if columns_filter is not None:
            filter_format = [f.upper() for f in columns_filter]
        result: List[str] = []
        for name, column_metadata in columns.items():
            if name not in filter_format:
                alias = getattr(column_metadata, 'alias')
                if alias:
                    result.append(f'{name} AS {alias}')
                else:
                    result.append(name)
        return result

    def retrieve_columns_as_str(
        self, table_name: str, columns_filter: List[str] = None, deleted: bool = False
    ) -> str:
        return ', '.join(
            self.retrieve_columns_as_list(
                table_name, columns_filter=columns_filter, deleted=deleted
            )
        )

    def retrieve_query_metadata_columns_as_str(
        self, columns_filter: List[str] = None, deleted: bool = False
    ) -> str:
        return ', '.join(
            self.retrieve_columns_as_list(
                '__INTERNAL__', columns_filter=columns_filter, deleted=deleted
            )
        )

    def retrieve_query_metadata_columns_as_list(
        self, columns_filter: List[str] = None, deleted: bool = False
    ) -> List[str]:
        return self.retrieve_columns_as_list(
            '__INTERNAL__', columns_filter=columns_filter, deleted=deleted
        )

    def retrieve_column_by_name(
        self, table_name: str, column_name: str, deleted: bool = False
    ) -> ColumnMetadata:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to retrieve column for table {t_name}({table_name}) - Table does not exist'
            )
        return self.tables[t_name].retrieve_column_by_column_name(column_name, deleted)

    def retrieve_query_metadata_column_by_name(self, column_name: str) -> ColumnMetadata:
        return self.retrieve_column_by_name('__INTERNAL__', column_name)

    def retrieve_query_metadata_column_type_by_name(self, column_name: str) -> str:
        return (
            self.retrieve_column_by_name('__INTERNAL__', column_name).type
            if hasattr(self.retrieve_column_by_name('__INTERNAL__', column_name), 'type')
            else 'UNDEFINED'
        )

    def retrieve_columns_by_type(
        self, table_name: str, column_type: str, deleted: bool = False
    ) -> Dict[str, ColumnMetadata]:
        t_name = table_name.upper()
        if t_name not in self.tables:
            raise MetadataError(
                f'Impossible to retrieve column for table {t_name}({table_name}) - Table does not exist'
            )
        return self.tables[t_name].retrieve_column_by_type(column_type, deleted)

    def retrieve_query_metadata_column_by_type(self, column_type: str) -> Dict[str, ColumnMetadata]:
        return self.retrieve_columns_by_type('__INTERNAL__', column_type)

    def cast_column_to_string(self, columns: Dict[str, ColumnMetadata]):
        return ', '.join([cname for cname, column in columns])

    def update_query_metadata_column_names_with_alias(self):
        for col in self.retrieve_query_metadata_columns().values():
            if col.alias:
                self.update_query_metadata_column_name(
                    column_name=col.name,
                    dest_column_name=col.alias,
                )
                self.remove_query_metadata_column_alias(col.alias)
        return self.retrieve_table('__INTERNAL__')

    def rename_union_columns(self):
        union_columns = self.retrieve_query_metadata_columns()
        [
            self.update_query_metadata_column_name(cname, cname.split(' ')[-1])
            for cname in union_columns.keys()
            if 'NULL' in cname
        ]
        return self



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

_SECOND = 1
_MINUTE = _SECOND * 60
_HOUR = _MINUTE * 60
_DAY = _HOUR * 24

DURATIONS_IN_SECOND = {'seconds': _SECOND, 'minutes': _MINUTE, 'hours': _HOUR, 'days': _DAY}


class DurationStep(BaseStep):
    name: str = Field('duration', const=True)
    new_column_name: ColumnName = Field(alias='newColumnName')
    start_date_column: ColumnName = Field(alias='startDateColumn')
    end_date_column: ColumnName = Field(alias='endDateColumn')
    duration_in: Literal['seconds', 'minutes', 'hours', 'days'] = Field(alias='durationIn')


class DurationStepWithVariable(DurationStep, StepWithVariablesMixin):
    duration_in: TemplatedVariable = Field(alias='durationIn')



import numpy as np
from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.statistics import DUMB_GROUPBY_COLUMN_NAME, StatisticsStep


def statistic_to_pandas_method(s: str) -> str:
    if s == 'average':
        return 'mean'
    elif s == 'variance':
        return 'var'
    elif s == 'standard deviation':
        return 'std'
    return s


def percentile(nth, order):
    def f(serie):
        return np.percentile(serie, nth / order * 100)

    return f


def execute_statistics(
    step: StatisticsStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    groupby_columns = step.groupby_columns
    if not step.groupby_columns:
        # Just create a temporary dumb column with '1' everywhere
        # so we can group by this column:
        groupby_columns = [DUMB_GROUPBY_COLUMN_NAME]
        df = df.assign(**{DUMB_GROUPBY_COLUMN_NAME: 1})

    stat_funcs = [statistic_to_pandas_method(s) for s in step.statistics]
    stat_funcs += [percentile(q.nth, q.order) for q in step.quantiles]
    result = df.groupby(groupby_columns)[step.column].agg(stat_funcs).reset_index()

    # Now, just fix the column names:
    quantile_names = [q.label or f'{q.nth}-th {q.order}-quantile' for q in step.quantiles]
    column_names = groupby_columns + step.statistics + quantile_names  # type: ignore
    result.columns = column_names

    if not step.groupby_columns:
        # Remove the temporary column:
        del result[DUMB_GROUPBY_COLUMN_NAME]

    return result



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    apply_condition,
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import IfthenelseStep
from weaverbird.pipeline.steps.ifthenelse import IfThenElse


def recursively_convert_nested_condition(step: IfthenelseStep, composed_query: str) -> str:
    if not hasattr(step, 'else_value'):
        return str(step)

    if type(step.else_value) != IfThenElse:
        return (
            f"""IFF({apply_condition(step.condition, composed_query)}, """
            f"""{step.then}, {step.else_value})"""
        )
    else:
        composed_query += (
            f"""IFF({apply_condition(step.condition, composed_query)}, {step.then}, """
            f"""{recursively_convert_nested_condition(step.else_value, composed_query)})"""
        )

    return composed_query.replace(',),', '),')


def translate_ifthenelse(
    step: IfthenelseStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'IFTHENELSE_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    composed_query: str = ''
    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=[step.new_column]
    )

    composed_query = (
        f"""{recursively_convert_nested_condition(step, composed_query).replace('"', "'")}"""
        f""" AS {step.new_column.upper()}"""
    )
    if completed_fields:
        composed_query = f', {composed_query}'

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}{composed_query}"""
        f""" FROM {query.query_name})""",
    )

    query.metadata_manager.add_query_metadata_column(step.new_column, 'str')

    new_query.selection_query = build_selection_query(
        query.metadata_manager.retrieve_query_metadata_columns(), query_name
    )
    new_query.metadata_manager = query.metadata_manager

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import RenameStep


def execute_rename(
    step: RenameStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    return df.rename(columns=dict(step.to_rename))



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import ConvertStep


def translate_convert(
    step: ConvertStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'CONVERT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step.columns: {step.columns}\n'
        f'step.data_type: {step.data_type}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    to_cast = []
    for col in step.columns:
        retrieved_col_type = query.metadata_manager.retrieve_query_metadata_column_type_by_name(
            column_name=col
        )
        if (
            retrieved_col_type == 'FLOAT' or retrieved_col_type == 'REAL'
        ) and step.data_type == 'integer':
            to_cast.append(f'TRUNCATE({col}) AS {col}')
        elif retrieved_col_type == 'TEXT' and step.data_type == 'integer':
            to_cast.append(f"CAST(SPLIT_PART({col}, '.', 0) AS {step.data_type}) AS {col}")
        elif (
            retrieved_col_type == 'TIMESTAMP_NTZ' or retrieved_col_type == 'DATE'
        ) and step.data_type == 'integer':
            to_cast.append(
                f"CAST(DATE_PART('EPOCH_MILLISECOND', TO_TIMESTAMP({col})) AS {step.data_type}) AS {col}"
            )
        else:
            to_cast.append(f'CAST({col} AS {step.data_type}) AS {col}')

    for c in step.columns:
        query.metadata_manager.update_query_metadata_column_type(c, step.data_type)

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=step.columns
    )

    compiled_query = ', '.join(to_cast)
    if len(completed_fields):
        compiled_query = f', {compiled_query}'
    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}{compiled_query}"""
        f""" FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import PercentageStep


def translate_percentage(
    step: PercentageStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'PERCENTAGE_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
    )

    new_column_name = f'{step.column}_PCT' if step.new_column_name is None else step.new_column_name
    group_query = (
        (f' GROUP BY {", ".join(step.group + [step.column])}' if len(step.group) > 0 else '')
        if step.group is not None
        else ''
    )

    select_fields = ""
    # added the new column
    for table in query.metadata_manager.tables:
        select_fields = (
            (
                ', '.join(step.group + [step.column])
                if len(step.group) > 0
                else ', '.join(query.metadata_manager.retrieve_columns_as_list(table))
            )
            if step.group is not None
            else ', '.join(query.metadata_manager.retrieve_columns_as_list(table))
        )

    for table in query.metadata_manager.tables:
        query.metadata_manager.add_table_column(table, new_column_name, 'float')

        if step.group is not None and len(step.group) > 0:
            for col in query.metadata_manager.retrieve_columns_as_list(table):
                if col not in (step.group + [step.column, new_column_name]):
                    query.metadata_manager.remove_query_metadata_column(col)

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {select_fields},"""
        f""" RATIO_TO_REPORT({step.column}) OVER () AS {new_column_name}"""
        f""" FROM {query.query_name}{group_query})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Tuple

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    first_last_query_string_with_group_and_granularity,
    generate_query_by_keeping_granularity,
    remove_metadatas_columns_from_query,
)
from weaverbird.backends.sql_translator.types import SQLQuery
from weaverbird.pipeline.steps import AggregateStep
from weaverbird.pipeline.steps.aggregate import Aggregation


def get_aggs_columns(step: AggregateStep, agg_type: str = "") -> list:
    """
    This method will return the first's/last's aggregations columns
    or normal aggregations depending on the agg_type
    default value is empty, that means it will return all agregations except first and last

    params:
    step: the aggregate step
    agg_type: the aggregate type (first/last)
    """

    def get_old_new_columns(aggregation: Aggregation):
        """
        From new/old columns this function will return a list of tuple
        from an aggregation
        """
        result = []
        # we loop to construct our first/last columns
        for col, new_col in zip(aggregation.columns, aggregation.new_columns):
            result.append((col, new_col))

        return result

    result = []
    for aggregation in step.aggregations:
        if agg_type == "":
            result += get_old_new_columns(aggregation)
        elif aggregation.agg_function == agg_type:
            result += get_old_new_columns(aggregation)

    return result


def build_first_or_last_aggregation(
    aggregated_string: str,
    first_last_string: str,
    query: SQLQuery,
    step: AggregateStep,
    query_name: str,
    new_as_columns=None,
) -> Tuple[SQLQuery, str]:
    """
    This method will help us build the first-last aggregation query

    params:
    aggregated_string: the aggregate string
    first_last_string: the first_last string
    query: the incoming sql-query
    step: the aggregate step
    """
    # extract first-last cols
    if new_as_columns is None:
        new_as_columns = []

    first_cols = get_aggs_columns(step, "first")
    last_cols = get_aggs_columns(step, "last")
    aggregate_cols = get_aggs_columns(step)

    query_string: str = ""

    def clean_none_aggregate_columns():
        """This pseudo method will remove all columns except the one provided"""
        return remove_metadatas_columns_from_query(
            query=query,
            array_cols=[f'{c[1]}' for c in aggregate_cols] + step.on + new_as_columns,
            first_last_string=query_string,
            first_or_last_aggregate=False,
        )

    # we add metadata aggregatescolumns
    [
        query.metadata_manager.add_query_metadata_column(target_col, "FLOAT")
        for target_col in [col[1] for col in aggregate_cols] + step.on
    ]

    # we build first/last aggregation query with a group by and granularity
    query, first_last_string = first_last_query_string_with_group_and_granularity(
        step=step,
        query=query,
        scope_cols=aggregate_cols,
    )

    if len(aggregated_string) and (len(first_cols) or len(last_cols)):
        group_by_sub_query: str = ""
        on_join_sub_query: str = ""
        if len(step.on):
            group_by_sub_query = f" ON {' AND '.join([f'A.{s}=F.{s}' for s in step.on])}"
            if step.keep_original_granularity:
                on_join_sub_query = (
                    f" AND (A.TO_REMOVE_{query_name} = F.TO_REMOVE_{query.query_name}_ALIAS)"
                )
            else:
                # we fresh the query and concatenate the previous query string
                query, query_string = clean_none_aggregate_columns()
        else:
            if step.keep_original_granularity:
                on_join_sub_query = (
                    f" ON (A.TO_REMOVE_{query_name} = F.TO_REMOVE_{query.query_name}_ALIAS)"
                )
            else:
                # we fresh the query and concatenate the previous query string
                query, query_string = clean_none_aggregate_columns()

        # we get our final query string
        query_string = (
            f"SELECT A.*, {', '.join([f'F.{c[1]}' for c in first_cols + last_cols])} FROM ({aggregated_string})"
            f" A INNER JOIN ({first_last_string}) F{group_by_sub_query}{on_join_sub_query}"
        )

    elif len(aggregated_string):
        query_string = aggregated_string
    else:
        query_string = first_last_string

    return query, query_string


def prepare_aggregation_query(
    query_name: str,
    aggregated_cols: list,
    aggregated_string: str,
    query: SQLQuery,
    step: AggregateStep,
) -> Tuple[SQLQuery, str, list]:
    new_as_columns: list = []

    for agg in step.aggregations:  # TODO the front should restrict - usage in column names
        agg.new_columns = [x.replace('-', '_').replace(' ', '_') for x in agg.new_columns]

    for aggregation in [
        aggregation
        for aggregation in step.aggregations
        if aggregation.agg_function not in ['first', 'last']
    ]:

        for col, new_col in zip(aggregation.columns, aggregation.new_columns):
            if aggregation.agg_function == 'count distinct':
                aggregated_cols.append(f'COUNT(DISTINCT {col}) AS {new_col}')
            else:
                aggregated_cols.append(f'{aggregation.agg_function.upper()}({col}) AS {new_col}')

    if len(step.on) and len(aggregated_cols):
        if step.keep_original_granularity:
            # we generate the query by keeping granularity
            # and update the query metadatas for new as columns
            query, aggregated_string, new_as_columns = generate_query_by_keeping_granularity(
                query=query,
                group_by=step.on,
                aggregated_cols=aggregated_cols,
                current_step_name=query_name,
            )

        else:
            aggregated_string = (
                f"SELECT * FROM (SELECT {', '.join(step.on + aggregated_cols)}"
                f" FROM {query.query_name} GROUP BY {', '.join(step.on)} ORDER BY {', '.join(step.on)})"
                f" {query.query_name}_ALIAS"
            )
            # we fresh the query and concatenate the previous query string
            query, _ = remove_metadatas_columns_from_query(
                query,
                step.on + [c.split(" AS ")[1] for c in aggregated_cols],
                "",
                False,
            )
    elif len(aggregated_cols):
        aggregated_string = f"SELECT {', '.join(aggregated_cols)} FROM {query.query_name}"
    return query, aggregated_string, new_as_columns



from pandas import Series, to_datetime, to_numeric


def cast_to_float(s: Series) -> Series:
    return to_numeric(s, errors='coerce')  # cast errors will result in NaN values


def cast_to_int(s: Series) -> Series:
    s_float = cast_to_float(s)
    # Since .astype(int) would fail on NaN values, treat them separately:
    s_integer = s_float[s_float.notna()].astype(int)
    s_nan = s_float[s_float.isna()]
    # Pandas operate on NanoSecond Timestamp but we want to output MilliSecond Timestamps
    if str(s.dtype).startswith('datetime64'):
        s_integer = s_integer // 10 ** 6
    return s_integer.append(s_nan).sort_index()


def cast_to_str(s: Series) -> Series:
    return s.astype(str)


def cast_to_datetime(s: Series) -> Series:
    is_casting_timestamps = str(s.dtype).startswith('int')
    return to_datetime(
        s, errors='coerce', unit='ms' if is_casting_timestamps else None
    )  # cast errors will result in NaT values


def cast_to_bool(s: Series) -> Series:
    return s.astype(bool)



from typing import Any, Dict, List, Union

from pydantic import BaseModel

from .steps import (
    AddMissingDatesStep,
    AddMissingDatesStepWithVariables,
    AggregateStep,
    AggregateStepWithVariables,
    AppendStep,
    AppendStepWithVariable,
    ArgmaxStep,
    ArgmaxStepWithVariable,
    ArgminStep,
    ArgminStepWithVariable,
    CompareTextStep,
    CompareTextStepWithVariables,
    ConcatenateStep,
    ConcatenateStepWithVariable,
    ConvertStep,
    CumSumStep,
    CumSumStepWithVariable,
    CustomSqlStep,
    DateExtractStep,
    DateExtractStepWithVariable,
    DeleteStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    DurationStepWithVariable,
    EvolutionStep,
    EvolutionStepWithVariable,
    FillnaStep,
    FillnaStepWithVariable,
    FilterStep,
    FilterStepWithVariables,
    FormulaStep,
    FormulaStepWithVariable,
    FromdateStep,
    IfthenelseStep,
    IfThenElseStepWithVariables,
    JoinStep,
    JoinStepWithVariable,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    PivotStepWithVariable,
    RankStep,
    RankStepWithVariable,
    RenameStep,
    RenameStepWithVariable,
    ReplaceStep,
    ReplaceStepWithVariable,
    RollupStep,
    RollupStepWithVariable,
    SelectStep,
    SortStep,
    SplitStep,
    SplitStepWithVariable,
    StatisticsStep,
    SubstringStep,
    TextStep,
    TextStepWithVariable,
    ToDateStep,
    TopStep,
    TopStepWithVariables,
    TotalsStep,
    TotalsStepWithVariable,
    TrimStep,
    UniqueGroupsStep,
    UniqueGroupsStepWithVariable,
    UnpivotStep,
    UnpivotStepWithVariable,
    UppercaseStep,
    WaterfallStep,
    WaterfallStepWithVariable,
)

PipelineStep = Union[
    AddMissingDatesStep,
    AggregateStep,
    AppendStep,
    ArgmaxStep,
    ArgminStep,
    CompareTextStep,
    ConcatenateStep,
    ConvertStep,
    CumSumStep,
    CustomSqlStep,
    DateExtractStep,
    DeleteStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    EvolutionStep,
    FillnaStep,
    FilterStep,
    FormulaStep,
    FromdateStep,
    FromdateStep,
    IfthenelseStep,
    JoinStep,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    RankStep,
    RenameStep,
    ReplaceStep,
    RollupStep,
    SelectStep,
    SortStep,
    SplitStep,
    StatisticsStep,
    SubstringStep,
    TextStep,
    ToDateStep,
    TopStep,
    TotalsStep,
    TrimStep,
    UniqueGroupsStep,
    UnpivotStep,
    UppercaseStep,
    WaterfallStep,
]


class Pipeline(BaseModel):
    steps: List[PipelineStep]

    def dict(self, *, exclude_none: bool = True, **kwargs) -> Dict:
        return super().dict(exclude_none=True, **kwargs)


PipelineStepWithVariables = Union[
    AddMissingDatesStepWithVariables,
    AggregateStepWithVariables,
    AppendStepWithVariable,
    ArgmaxStepWithVariable,
    ArgminStepWithVariable,
    CompareTextStepWithVariables,
    ConcatenateStepWithVariable,
    CumSumStepWithVariable,
    DateExtractStepWithVariable,
    DurationStepWithVariable,
    EvolutionStepWithVariable,
    FillnaStepWithVariable,
    FilterStepWithVariables,
    FormulaStepWithVariable,
    IfThenElseStepWithVariables,
    JoinStepWithVariable,
    PivotStepWithVariable,
    RankStepWithVariable,
    RenameStepWithVariable,
    ReplaceStepWithVariable,
    RollupStepWithVariable,
    SplitStepWithVariable,
    TextStepWithVariable,
    TopStepWithVariables,
    TotalsStepWithVariable,
    UniqueGroupsStepWithVariable,
    UnpivotStepWithVariable,
    WaterfallStepWithVariable,
]


class PipelineWithVariables(BaseModel):
    steps: List[Union[PipelineStepWithVariables, PipelineStep]]

    def render(self, variables: Dict[str, Any], renderer) -> Pipeline:
        # TODO it must be more efficient to render the full pipeline once
        steps_rendered = [
            step.render(variables, renderer) if hasattr(step, 'render') else step for step in self.steps  # type: ignore
        ]
        return Pipeline(steps=steps_rendered)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import SplitStep


def execute_split(
    step: SplitStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    all_new_col = {}
    new_serie = df[step.column].str.split(step.delimiter)
    for i in range(step.number_cols_to_keep):
        all_new_col[f'{step.column}_{i + 1}'] = new_serie.str[i]
    return df.assign(**all_new_col)



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import SplitStep


def translate_split(
    step: SplitStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f"SPLIT_STEP_{index}"

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    # at least one col to keep
    step.number_cols_to_keep = (
        1
        if step.number_cols_to_keep is None or step.number_cols_to_keep == 0
        else step.number_cols_to_keep
    )

    # we should escape weird quotes on the delimiter
    # to prevent the sql query fail
    step.delimiter = step.delimiter.replace('"', "'").replace("'", "\\'")

    # We complete fields
    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str()

    # We construct the split query here
    # using SPLIT_PART
    split_query = ", ".join(
        [
            f"SPLIT_PART({step.column}, '{step.delimiter}', {delimiter_count + 1}) AS {step.column}_{delimiter_count + 1}"
            for delimiter_count in range(0, step.number_cols_to_keep)
        ]
    )
    # We add the metadata column
    [
        query.metadata_manager.add_query_metadata_column(
            f"{step.column}_{delimiter_count + 1}", "text"
        )
        for delimiter_count in range(0, step.number_cols_to_keep)
    ]

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}, {split_query}"""
        f""" FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



import json
import logging
from typing import Tuple

from pandas import DataFrame
from pandas.io.json import build_table_schema

from weaverbird.backends.pandas_executor.steps import steps_executors
from weaverbird.backends.pandas_executor.types import (
    DomainRetriever,
    PipelineExecutionReport,
    StepExecutionReport,
)
from weaverbird.pipeline import Pipeline, PipelineStep
from weaverbird.utils import StopWatch, convert_size

logger = logging.getLogger(__name__)


def execute_pipeline(
    pipeline: Pipeline, domain_retriever: DomainRetriever
) -> Tuple[DataFrame, PipelineExecutionReport]:
    """
    The main function of the module. Execute a pipeline and returns the result as a pandas DataFrame.

    Its domain retriever will be use to retrieve data needed from `domains` steps.
    """
    # TODO validate the pipeline, e.g. the first step should always be a domain step
    # validate_pipeline()
    steps = pipeline.steps
    df = None
    stopwatch = StopWatch()
    step_reports = []
    for index, step in enumerate(steps):
        try:
            with stopwatch:
                df = steps_executors[step.name](
                    step,
                    df,
                    domain_retriever=domain_retriever,
                    execute_pipeline=execute_pipeline,
                )
            logger.info(
                f'step {step} used {convert_size(df.memory_usage().sum())}, took {stopwatch.interval}s to execute'
            )
            step_reports.append(
                StepExecutionReport(
                    step_index=index,
                    memory_used_in_bytes=df.memory_usage().sum(),
                    time_spent_in_ms=int(stopwatch.interval * 1000),
                )
            )
        except Exception as e:
            raise PipelineExecutionFailure(step, index, e) from e
    return df, PipelineExecutionReport(steps_reports=step_reports)


def preview_pipeline(
    pipeline: Pipeline, domain_retriever: DomainRetriever, limit: int = 50, offset: int = 0
) -> str:
    """
    Execute a pipeline but returns only a slice of the results, determined by `limit` and `offset` parameters, as JSON.

    Return format follows the 'table' JSON table schema used by pandas (see
    https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#orient-options), with a few addition related to
    pagination.

    Note: it's required to use pandas `to_json` methods, as it convert NaN and dates to an appropriate format.
    """
    df, _ = execute_pipeline(pipeline, domain_retriever)
    return json.dumps(
        {
            'schema': build_table_schema(df, index=False),
            'offset': offset,
            'limit': limit,
            'total': df.shape[0],
            'data': json.loads(df[offset : offset + limit].to_json(orient='records')),
        }
    )


class PipelineExecutionFailure(Exception):
    """Raised when a error happens during the execution of the pipeline"""

    def __init__(self, step: PipelineStep, index: int, original_exception: Exception):
        self.step = step
        self.index = index
        self.original_exception = original_exception
        self.message = f'Step #{index + 1} ({step.name}) failed: {original_exception}'
        self.details = {'index': index, 'message': self.message}



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.utils.combination import PipelineOrDomainName


def resolve_pipeline_for_combination(
    pipeline: PipelineOrDomainName,
    domain_retriever: DomainRetriever,
    pipeline_executor: PipelineExecutor,
) -> DataFrame:
    """
    Combined pipelines can be either single domains (str), or complete pipeline (list of steps)
    """
    from weaverbird.pipeline import Pipeline

    if isinstance(pipeline, str):
        return domain_retriever(pipeline)
    else:
        # NOTE execution report of the sub-pipeline is discarded
        return pipeline_executor(Pipeline(steps=pipeline), domain_retriever)[0]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import UppercaseStep


def translate_uppercase(
    step: UppercaseStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'UPPERCASE_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step.name: {step.name}\n'
        f'step.column: {step.column}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=[step.column]
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields},"""
        f""" UPPER({step.column}) AS {step.column.upper()}"""
        f""" FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Dict

from ..types import StepExecutor
from .addmissingdates import execute_addmissingdates
from .aggregate import execute_aggregate
from .append import execute_append
from .argmax import execute_argmax
from .argmin import execute_argmin
from .comparetext import execute_comparetext
from .concatenate import execute_concatenate
from .convert import execute_convert
from .cumsum import execute_cumsum
from .date_extract import execute_date_extract
from .delete import execute_delete
from .domain import execute_domain
from .duplicate import execute_duplicate
from .duration import execute_duration
from .evolution import execute_evolution
from .fillna import execute_fillna
from .filter import execute_filter
from .formula import execute_formula
from .fromdate import execute_fromdate
from .ifthenelse import execute_ifthenelse
from .join import execute_join
from .lowercase import execute_lowercase
from .moving_average import execute_moving_average
from .percentage import execute_percentage
from .pivot import execute_pivot
from .rank import execute_rank
from .rename import execute_rename
from .replace import execute_replace
from .rollup import execute_rollup
from .select import execute_select
from .sort import execute_sort
from .split import execute_split
from .statistics import execute_statistics
from .substring import execute_substring
from .text import execute_text
from .todate import execute_todate
from .top import execute_top
from .totals import execute_totals
from .trim import execute_trim
from .uniquegroups import execute_uniquegroups
from .unpivot import execute_unpivot
from .uppercase import execute_uppercase
from .waterfall import execute_waterfall

steps_executors: Dict[str, StepExecutor] = {
    'addmissingdates': execute_addmissingdates,
    'aggregate': execute_aggregate,
    'append': execute_append,  # type: ignore
    'argmax': execute_argmax,
    'argmin': execute_argmin,
    'comparetext': execute_comparetext,
    'concatenate': execute_concatenate,
    'convert': execute_convert,
    'cumsum': execute_cumsum,
    'dateextract': execute_date_extract,
    'delete': execute_delete,
    'domain': execute_domain,  # type: ignore
    'duplicate': execute_duplicate,
    'duration': execute_duration,
    'evolution': execute_evolution,
    'fillna': execute_fillna,
    'filter': execute_filter,
    'formula': execute_formula,
    'fromdate': execute_fromdate,
    'ifthenelse': execute_ifthenelse,
    'join': execute_join,  # type: ignore
    'lowercase': execute_lowercase,
    'movingaverage': execute_moving_average,
    'percentage': execute_percentage,
    'pivot': execute_pivot,
    'rank': execute_rank,
    'rename': execute_rename,
    'replace': execute_replace,
    'rollup': execute_rollup,
    'select': execute_select,
    'sort': execute_sort,
    'split': execute_split,
    'statistics': execute_statistics,
    'substring': execute_substring,
    'text': execute_text,
    'todate': execute_todate,
    'top': execute_top,
    'totals': execute_totals,
    'trim': execute_trim,
    'uniquegroups': execute_uniquegroups,
    'unpivot': execute_unpivot,
    'uppercase': execute_uppercase,
    'waterfall': execute_waterfall,
}



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import UnpivotStep


def translate_unpivot(
    step: UnpivotStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'UNPIVOT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )
    unpivot_query = f"""SELECT {', '.join(step.keep + [step.unpivot_column_name.upper(),
                                                       step.value_column_name.upper()])}\
 FROM {query.query_name} UNPIVOT({step.value_column_name.upper()} FOR {step.unpivot_column_name.upper()} IN\
 {str(step.unpivot).replace('[', '(').replace(']', ')').replace("'", '')}"""

    transformed_query = f"""{query.transformed_query}, {query_name} AS ({unpivot_query}))"""

    unpivotted_value_column_type = (
        query.metadata_manager.retrieve_query_metadata_column_type_by_name(step.unpivot[0])
    )
    query.metadata_manager.remove_query_metadata_columns(
        [
            c
            for c in query.metadata_manager.retrieve_query_metadata_columns().keys()
            if c not in step.keep
        ]
    )
    query.metadata_manager.add_query_metadata_column(
        column_name=step.unpivot_column_name, column_type='text'
    )
    query.metadata_manager.add_query_metadata_column(
        column_name=step.value_column_name,
        # We infer that type of value column will be the same as first unpivoted column's type.
        # If the provided columns to unpivot have mixed types the unpivot will fail.
        column_type=unpivotted_value_column_type,
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import List, Optional

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class CumSumStep(BaseStep):
    name = Field('cumsum', const=True)
    value_column: ColumnName = Field(..., alias='valueColumn')
    reference_column: ColumnName = Field(..., alias='referenceColumn')
    groupby: Optional[List[ColumnName]]
    new_column: Optional[ColumnName] = Field(None, alias='newColumn')


class CumSumStepWithVariable(CumSumStep, StepWithVariablesMixin):
    ...



import re
from typing import Dict, List, Tuple

from weaverbird.backends.sql_translator.metadata import ColumnMetadata, SqlQueryMetadataManager
from weaverbird.backends.sql_translator.types import SQLQuery
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.steps import AggregateStep

SQL_COMPARISON_OPERATORS = {
    'eq': '=',
    'ne': '!=',
    'lt': '<',
    'le': '<=',
    'gt': '>',
    'ge': '>=',
}

SQL_NULLITY_OPERATORS = {
    'isnull': 'IS NULL',
    'notnull': 'IS NOT NULL',
}

SQL_MATCH_OPERATORS = {
    'matches': 'RLIKE',
    'notmatches': 'NOT RLIKE',
}

SQL_INCLUSION_OPERATORS = {
    'in': 'IN',
    'nin': 'NOT IN',
}


def apply_condition(condition: Condition, query: str) -> str:
    if isinstance(condition, ComparisonCondition):
        try:
            float(condition.value)
            query += f'{condition.column} {SQL_COMPARISON_OPERATORS[condition.operator]} {condition.value}'
        except ValueError:
            # just to escape single quotes from crashing the snowflakeSQL query
            if type(condition.value) == str:
                condition.value = sanitize_input(condition.value)
            query += f"{condition.column} {SQL_COMPARISON_OPERATORS[condition.operator]} '{condition.value}'"
    elif isinstance(condition, NullCondition):
        query += f'{condition.column} {SQL_NULLITY_OPERATORS[condition.operator]}'
    elif isinstance(condition, MatchCondition):
        # just to escape single quotes from crashing the snowflakeSQL query
        if type(condition.value) == str:
            condition.value = sanitize_input(condition.value)
        query += f"{condition.column} {SQL_MATCH_OPERATORS[condition.operator]} '{condition.value}'"
    elif isinstance(condition, InclusionCondition):
        query += f'{condition.column} {SQL_INCLUSION_OPERATORS[condition.operator]} {str(tuple(condition.value))}'
    elif isinstance(condition, ConditionComboAnd):
        query = apply_condition(condition.and_[0], query)
        for c in condition.and_[1:]:
            query = apply_condition(c, f'{query} AND ')
    elif isinstance(condition, ConditionComboOr):
        query = apply_condition(condition.or_[0], query)
        for c in condition.or_[1:]:
            query = apply_condition(c, f'{query} OR ')
    else:
        raise NotImplementedError('Only comparison conditions are implemented')
    return query


def build_selection_query(query_metadata: Dict[str, ColumnMetadata], query_name) -> str:
    names = []
    for _, metadata in query_metadata.items():
        alias = getattr(metadata, 'alias')
        if alias:
            names.append(alias)
        else:
            names.append(getattr(metadata, 'name'))
    return f"SELECT {', '.join(names)} FROM {query_name}"


def first_last_query_string_with_group_and_granularity(
    step: AggregateStep,
    query: SQLQuery,
    scope_cols: list,
) -> Tuple[SQLQuery, str]:
    """
    This function will... depending on the group by of the aggregate and the granularity conservation, generate
    the appropriate final query and clean/update the query's metadata columns

    params:
    query: the sqlquery object
    step: The current aggregate step
    select_array_cols: the selection columns in the end_query
    array_cols: the order by's columns
    scope_cols: he scope of our process, for example: firsts_cols or lasts_cols...
    """

    if len(step.on):
        # depending on the granularity keep parameter
        # we should remove unnecessary columns
        if step.keep_original_granularity:
            groupby_alias = ', '.join([f'{cc} AS X_{cc}' for cc in step.on])
            end_query = (
                f"SELECT {groupby_alias},{', '.join([f'{col} AS {new_col}' for (col, new_col) in scope_cols]+['ROW_NUMBER()'])}"
                f" OVER (PARTITION BY {', '.join(step.on)}"
                f" ORDER BY {', '.join([f'{c[0]}' for c in scope_cols])}) AS R FROM {query.query_name} QUALIFY R = 1"
            )
            # ['X.VALUE__first=Z.VALUE__first', 'X.TIME_first=Z.TIME_first']
            final_end_query_string = (
                f"SELECT * FROM ({end_query}) X "
                f"INNER JOIN (SELECT *,ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) AS TO_REMOVE_{query.query_name}_ALIAS"
                f" FROM {query.query_name}) Z"
                f" ON {' AND '.join([f'X.X_{s}=Z.{s}' for s in step.on])} ORDER BY TO_REMOVE_{query.query_name}_ALIAS"
            )
        else:
            # the difference on the  if col != new_col after the loop
            end_query = (
                f"SELECT *,"
                f"{', '.join([f'{col} AS {new_col}' for (col, new_col) in scope_cols if col != new_col] + ['ROW_NUMBER()'])}"
                f" OVER (PARTITION BY {', '.join(step.on)}"
                f" ORDER BY {', '.join([f'{c[0]}' for c in scope_cols])}) AS R FROM {query.query_name} QUALIFY R = 1"
            )
            # we fresh an concatenate the final first_last_string
            query, final_end_query_string = remove_metadatas_columns_from_query(
                query, [c[1] for c in scope_cols] + step.on, end_query
            )
    else:
        # depending on the granularity keep parameter
        # we should remove unnecessary columns
        if step.keep_original_granularity:
            end_query = (
                f"SELECT "
                f"{', '.join(step.on + [f'{col} AS {new_col}' for (col, new_col) in scope_cols] + ['ROW_NUMBER()'])}"
                "OVER ("
                f"ORDER BY {', '.join([f'{c[0]}' for c in scope_cols])}) AS R FROM {query.query_name} QUALIFY R = 1"
            )
            final_end_query_string = (
                f"SELECT * FROM ({end_query}) X INNER JOIN {query.query_name} Z "
            )
        else:
            end_query = (
                f"SELECT *,"
                f"{', '.join(step.on + [f'{col} AS {new_col}' for (col, new_col) in scope_cols if col != new_col] + ['ROW_NUMBER()'])} OVER ("
                f"ORDER BY {', '.join([f'{c[0]}' for c in scope_cols])}) AS R FROM {query.query_name} QUALIFY R = 1"
            )

            # we fresh an concatenate the final first_last_string
            query, final_end_query_string = remove_metadatas_columns_from_query(
                query, [c[1] for c in scope_cols], end_query
            )

    return query, final_end_query_string


def remove_metadatas_columns_from_query(
    query: SQLQuery, array_cols: list, first_last_string: str, first_or_last_aggregate: bool = True
) -> Tuple[SQLQuery, str]:
    """
    For the given query, this function will remove metadata columns if its not on a given list
    then concatenate the join on array_cols list to the final query

    params:
    query : the given query
    array_cols : the list of columns we want to keep in the metadatas
    str_query : the end query
    """

    # we loop on tables and add new columns and get back the fresh query
    # we add metadata columns
    [
        query.metadata_manager.remove_query_metadata_column(col)
        for col in list(query.metadata_manager.retrieve_query_metadata_columns().keys())
        if col.upper() not in [c.upper() for c in array_cols]
    ]

    if first_or_last_aggregate:
        first_last_string = f"SELECT {', '.join(array_cols)} FROM ({first_last_string})"

    return query, first_last_string


def generate_query_by_keeping_granularity(
    query: SQLQuery,
    group_by: list,
    current_step_name: str,
    query_to_complete: str = "",
    aggregated_cols=None,
    group_by_except_target_columns=None,
) -> Tuple[SQLQuery, str, list]:
    """
    On some steps, when we do the Group By we will need to keep the granularity of
    all our precedents columns, this method will do that operation but with an innerjoin on the precedent dataset
    then order by those elements in the group by list/array
    For example :
    SELECT * FROM (
        WITH SELECT_STEP_0 AS (
            SELECT Price, Name, Category FROM products
        ), AGGREGATE_STEP_1 AS (
            SELECT SUM(Price) AS PRICE_SUM, Category FROM SELECT_STEP_0 GROUP BY Category
        )  A INNER JOIN (SELECT * FROM SELECT_STEP_0) B
                ON A.Category = B.Category
                ORDER BY A.Category
    ) SELECT * FROM AGGREGATE_STEP_1

    params:
    group_by: the group by list
    prev_step_name: the previous step name (usually query.query_name)
    current_step_name: the current step name (usually query_name)

    it's supposed to return :
    """
    # We build the group by query part
    if group_by_except_target_columns is None:
        group_by_except_target_columns = []
    if aggregated_cols is None:
        aggregated_cols = []

    group_by_query: str = ""
    on_query: str = ""
    sub_select_query: str = ""

    for index, gb in enumerate(group_by):
        # we create a subfield containing a the alias name for the current column and the index
        sub_field = f"{gb}_ALIAS_{index}"

        # The sub select query
        sub_select_query += ("" if index == 0 else ", ") + f"{gb} AS {sub_field}"

        # The ON query re-group
        on_query += (
            "" if index == 0 else " AND "
        ) + f"({sub_field} = {query.query_name}_ALIAS.{gb})"

        # The GROUP BY query
        group_by_query += ('GROUP BY ' if index == 0 else ', ') + sub_field

    new_as_columns: list = []
    for index, ag in enumerate(aggregated_cols):
        # the aggregate as word
        as_ag = ag.split(" AS ")[1]

        new_as_columns.append(as_ag)

        # The sub select query
        sub_select_query += f", {ag.split(' AS ')[0]} AS {as_ag}"

        # we apend to the array of metadata
        query.metadata_manager.add_query_metadata_column(as_ag, "float")

    return (
        query,
        (
            f"SELECT * FROM (SELECT {sub_select_query}{query_to_complete}"
            f" FROM {query.query_name} {group_by_query}) {current_step_name}_ALIAS"
            f" INNER JOIN (SELECT *, ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) AS TO_REMOVE_{current_step_name}"
            f" FROM {query.query_name})"
            f" {query.query_name}_ALIAS ON ({on_query}) ORDER BY TO_REMOVE_{current_step_name}"
        ),
        new_as_columns,
    )


def snowflake_date_format(input_format: str) -> str:
    """
    This method will format the standard sql format for dates into snowflake sql one
    """
    # we escape quotes here and construct our format
    # for a valid snowflake date format
    input_format = (
        None
        if input_format is None or input_format == ''
        else input_format.replace('"', '')
        .replace("'", '')
        .replace('%b', 'MON')
        .replace('%B', 'MMMM')
        .replace('%y', 'YYYY')
        .replace('%Y', 'YYYY')
        .replace('%M', 'MM')
        .replace('%m', 'MM')
        .replace('%D', 'DD')
        .replace('%d', 'DD')
    )
    input_format = '' if input_format is None or input_format == '' else f", '{input_format}'"

    return input_format


def handle_zero_division(formula: str) -> str:
    """
    Use regular expression replacement to detect '/' or '%' in formulas
    and substitute a 0 divisor by NULL
    For example in '1 % BLAAA / "BLA    BLA"'
    this r'(?<=%)\s*(\w+)' captures BLAAA and this r' NULLIF(\1, 0)' replaces it
    by NULLIF(BLAAA, 0).
    this r'(?<=/)\s*(\"?.*\"?)' captures "BLA    BLA" and this r'NULLIFF(\2, 0)' replaces it
    by NULLIF("BLA    BLA", 0).
    """
    if '/' not in formula and '%' not in formula:
        return formula
    if '/' in formula:
        formula = re.sub(r'(?<=/)\s*(\w+)|(?<=/)\s*(\"?.*\"?)', r' NULLIF(\1\2, 0)', formula)
    if '%' in formula:
        formula = re.sub(r'(?<=%)\s*(\w+)|(?<=%)\s*(\"?.*\"?)', r' NULLIF(\1\2, 0)', formula)
    return formula


def build_union_query(
    query_metadata_manager: SqlQueryMetadataManager,
    current_query_name: str,
    appended_tables_name: List[str],
) -> str:
    new_query = f'SELECT {query_metadata_manager.retrieve_query_metadata_columns_as_str()} FROM {current_query_name}'
    max_column_number = len(query_metadata_manager.retrieve_query_metadata_columns_as_list())
    for t in appended_tables_name:
        table_columns = query_metadata_manager.retrieve_columns_as_list(t)
        missing_columns = max_column_number - len(table_columns)
        all_columns = table_columns + ['NULL'] * missing_columns
        new_query += f" UNION ALL SELECT {', '.join(all_columns)} FROM {t}"
    return new_query


def get_query_for_date_extract(
    date_type: str,
    target_column: str,
    new_column: str,
) -> str:
    """
    This method will get as input the date type and return a query with
    the appropriate function of that date type on snowflake, it can be a simple function or a whole expression


    """
    if date_type.lower() in [
        "year",
        "month",
        "day",
        "hour",
        "minutes",
        "seconds",
        "dayofweek",
        "dayofweekiso",
        "dayofyear",
        "week",
        "weekiso",
        "quarter",
        "yearofweek",
        "yearofweekiso",
    ]:
        date_type = date_type[:-1] if date_type in ["seconds", "minutes"] else date_type
        return f"EXTRACT({date_type.lower()} from to_timestamp({target_column})) AS {new_column}"
    else:
        appropriate_func = {
            "milliseconds": "DATE_TRUNC(millisecond, to_timestamp(____target____))",
            "isoYear": "YEAROFWEEKISO(to_timestamp(____target____))",
            "isoWeek": "WEEKISO(to_timestamp(____target____))",
            "isoDayOfWeek": "DAYOFWEEKISO(to_timestamp(____target____))",
            "firstDayOfYear": "TO_TIMESTAMP_NTZ(DATE_TRUNC(year, to_timestamp(____target____)))",
            "firstDayOfMonth": "TO_TIMESTAMP_NTZ(DATE_TRUNC(month, to_timestamp(____target____)))",
            "firstDayOfWeek": "TO_TIMESTAMP_NTZ(DATE_TRUNC(week, to_timestamp(____target____)))",
            "firstDayOfQuarter": "TO_TIMESTAMP_NTZ(DATE_TRUNC(quarter, to_timestamp(____target____)))",
            "firstDayOfIsoWeek": "DAYOFWEEKISO(to_timestamp(____target____)) - DAYOFWEEKISO(to_timestamp("
            "____target____)) + 1",
            "previousDay": "to_timestamp(____target____) - interval '1 day'",
            "firstDayOfPreviousYear": "(to_timestamp(____target____) - interval '1 year') + interval '1 day'",
            "firstDayOfPreviousMonth": "(to_timestamp(____target____) - interval '2 month') + interval '1 day'",
            "firstDayOfPreviousWeek": "DAY(to_timestamp(____target____) - interval '1 week') - DAYOFWEEKISO("
            "to_timestamp(____target____)) + 1",
            "firstDayOfPreviousQuarter": "to_timestamp(____target____) - interval '1 quarter'",
            "firstDayOfPreviousIsoWeek": "DAYOFWEEKISO(to_timestamp(____target____) - interval '1 week') - "
            "DAYOFWEEKISO(to_timestamp(____target____)) + 1",
            "previousYear": "YEAR(to_timestamp(____target____) - interval '1 year')",
            "previousMonth": "MONTH(to_timestamp(____target____) - interval '1 month')",
            "previousWeek": "WEEK(to_timestamp(____target____) - interval '1 week')",
            "previousQuarter": "QUARTER(to_timestamp(____target____) - interval '1 quarter')",
            "previousIsoWeek": "WEEKISO(to_timestamp(____target____)) - 1",
        }

        return f"({appropriate_func[date_type].replace('____target____', target_column)}) AS {new_column}"


def sanitize_input(value: str) -> str:
    return value.replace('"', '\\"').replace("'", "\\'")


def sanitize_column_name(col: str) -> str:
    # see   https://docs.snowflake.com/en/sql-reference/identifiers-syntax.html
    if col[0].isdigit():
        return f"_{col[0]}{re.sub(r'[^0-9a-zA-Z_$]', '_', col[1:])}"
    else:
        return re.sub(r'[^0-9a-zA-Z_$]', '_', col)


def build_aggregated_columns(aggregations):
    aggregated_columns = []
    for aggregation in aggregations:
        for col, newcol in zip(aggregation.columns, aggregation.new_columns):
            aggregated_columns.append(
                f'{aggregation.agg_function.upper()}({col}) AS {newcol}'
                if aggregation.agg_function != 'count distinct'
                else f'COUNT(DISTINCT {col}) AS {newcol}'
            )
    return aggregated_columns


def build_hierarchical_columns_list(step):
    level_columns = ' '.join(
        [f"WHEN {c} IS NOT NULL THEN '{c.upper()}'" for c in step.hierarchy[::-1]]
    )
    level_columns = f"CASE {level_columns} ELSE '' END AS {step.level_col or 'LEVEL'}"
    label_columns = (
        f"""COALESCE({', '.join(step.hierarchy[::-1])}) AS {step.label_col or "LABEL"}"""
    )
    parent_columns = []
    for p, c in zip(step.hierarchy, step.hierarchy[1:]):
        parent_columns.append(f"""WHEN {step.level_col or "LEVEL"} = '{c}' THEN "{p}" """)
    parent_columns = (
        f"""CASE {''.join(parent_columns)}ELSE NULL END AS {step.parent_label_col or 'PARENT'}"""
    )

    aggregated_columns = build_aggregated_columns(step.aggregations)

    all_columns = (
        step.hierarchy
        + (step.groupby or [])
        + [label_columns, level_columns, parent_columns]
        + aggregated_columns
    )
    return ', '.join(all_columns)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import FromdateStep


def execute_fromdate(
    step: FromdateStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    formatted_time = df[step.column].dt.strftime(step.format)
    return df.assign(**{step.column: formatted_time})



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import SubstringStep


def execute_substring(
    step: SubstringStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column_name = step.new_column_name or f'{step.column}_SUBSTR'
    # Weaverbird indexes start at one
    start_index = step.start_index - 1
    # Weaverbird substring end_index is inclusive. in python, it is exclusive.
    # Therefore, no need to substract one
    end_index = step.end_index

    if end_index < 0:
        end_index += 1
    if start_index < 0:
        start_index += 1

    # there is no way to get the full string with a x[start:end] when end is negative.
    if end_index == 0:
        serie = df[step.column].str[start_index:]
    else:
        serie = df[step.column].str[start_index:end_index]

    return df.assign(**{new_column_name: serie})



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
    get_query_for_date_extract,
)
from weaverbird.backends.sql_translator.types import (
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import DateExtractStep


def translate_dateextract(
    step: DateExtractStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
) -> SQLQuery:
    query_name = f'DATEEXTRACT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
    )

    def new_column_name(d: str, idd: int) -> str:
        """
        A security method inside the translator to get a new column name base
        on the step.new_columns wetheir or not that is empty or full

        d : date info element
        idd: index of that date info in new columns array
        """
        return (
            (
                # if index is in step.new_columns
                step.new_columns[idd]
                if idd < len(step.new_columns)
                else f'{step.column}_{d.upper()}'
            )
            if len(step.new_columns) > 0
            else f'{step.column}_{d.upper()}'
        )

    to_extract_string: str = ", ".join(
        [
            get_query_for_date_extract(d, step.column, new_column_name(d, idd))
            for idd, d in enumerate(step.date_info)
            if query.metadata_manager.add_query_metadata_column(
                f"{new_column_name(d, idd)}", "date"
            )
        ]
    )

    select_fields = ""
    # added the new column
    for table in query.metadata_manager.tables:
        select_fields = ', '.join(query.metadata_manager.retrieve_columns_as_list(table))

    if len(select_fields) > 0:
        to_extract_string = ", " + to_extract_string

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {select_fields}{to_extract_string} FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query
